#!/usr/bin/env python
""" pentest.py

Copyright (C) 2016  Dale V. Patterson (wraith.wireless@yandex.com)

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

Redistribution and use in source and binary forms, with or without
modifications, are permitted provided that the following conditions are met:
 o Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 o Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 o Neither the name of the orginal author Dale V. Patterson nor the names of
    any contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

Example for setting up a wireless environment - must be done as root

"""

from __future__ import print_function  # python 2to3 compability
import argparse as ap                  # cli arg. parsing
import time                            # sleeping
import sys                             # cli exiting
import pyric                           # pyric error (and ecode EUNDEF)
import pyric.pyw as pyw                # for iw functionality
import pyric.utils.hardware as hw      # for chipset/driver
from pyric.utils.channels import rf2ch # rf to channel conversion

def execute(dev):
    print('Setting up...')
    # ensure dev is a wireless interfaces
    ifaces = pyw.interfaces()
    wifaces = pyw.winterfaces()
    if dev not in ifaces:
        print("Device {0} is not valid, use one of {1}".format(dev,ifaces))
        return
    elif dev not in wifaces:
        print("Device {0} is not wireless, use one of {1}".format(dev,wifaces))

    # get a Card & info for dev
    print("Regulatory Domain currently: ", pyw.regget())
    dinfo = pyw.devinfo(dev)
    card = dinfo['card']
    pinfo = pyw.phyinfo(card)
    driver = hw.ifdriver(card.dev)
    chipset = hw.ifchipset(driver)

    # bring the card down and change the mac
    pyw.down(card)
    pyw.macset(card,'00:03:93:57:54:46')

    # print details
    msg = "Using {0} currently in mode: {1}\n".format(card,dinfo['mode'])
    msg += "\tDriver: {0} Chipset: {1}\n".format(driver,chipset)
    if dinfo['mode'] == 'managed':
        msg += "\tcurrently on channel {0} width {1}\n".format(rf2ch(dinfo['RF']),
                                                               dinfo['CHW'])
    msg += "\tSupports modes {0}\n".format(pinfo['modes'])
    msg += "\tSupports commands {0}".format(pinfo['commands'])
    msg += "\thw addr {0}".format(pyw.macget(card))
    print(msg)

    # prepare a virtual interface named pent0 in monitor mode
    # delete all ifaces on the phy to avoid interference
    # bring the card up when down
    print('Preparing pent0 for monitor mode')
    pdev = 'pent0'
    pcard = pyw.devadd(card, pdev, 'monitor')
    for iface in pyw.ifaces(card):
        if iface[0].dev != pcard.dev:
            print("deleting {0} in mode {1}".format(iface[0],iface[1]))
            pyw.devdel(iface[0])
    pyw.up(pcard)
    print("Using", pcard)

    print("Setting channel to 6 NOHT")
    pyw.chset(pcard,6,None)
    msg = "Virtual interface {0} in monitor mode on ch 6".format(pcard)
    print(msg + ", using hwaddr: {0}".format(pyw.macget(pcard)))

    # DO stuff here
    try:
        print('Now ready to do stuff')
        print('For example, run wireshark to verify card is seeing all packets')
        print('Hit Ctrl-C to quit and restore')
        while True: time.sleep(1)
    except KeyboardInterrupt:
        pass

    # restore original
    print('Restoring', card, 'mode =', dinfo['mode'], 'mac =', dinfo['mac'])
    card = pyw.devadd(pcard,card.dev,dinfo['mode'])
    print('Deleting', pcard)
    pyw.devdel(pcard)
    pyw.macset(card,dinfo['mac'])
    pyw.up(card)
    print("card ", card, " restored")

if __name__ == '__main__':
    # create arg parser and parse command line args
    print("Wireless Pentest Environment using PyRIC v{0}".format(pyric.version))
    argp = ap.ArgumentParser(description="Wireless Pentest")
    argp.add_argument('-d','--dev',help="Pentesting Wireless Device")
    args = argp.parse_args()
    try:
        dname = args.dev
        if dname is None:
            print("usage: python pentest.py -d <dev>")
            sys.exit(0)
        else:
            execute(dname)
    except pyric.error as e:
        print(e)


#!/usr/bin/env python

""" sockios_h.py: definitions of the socket-level I/O control calls.

A port of sockios.h (and two constants from wireless.h) to python
/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		Definitions of the socket-level I/O control calls.
 *
 * Version:	@(#)sockios.h	1.0.2	03/09/93
 *
 * Authors:	Ross Biro
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

Copyright (C) 2016  Dale V. Patterson (wraith.wireless@yandex.com)

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

Redistribution and use in source and binary forms, with or without modifications,
are permitted provided that the following conditions are met:
 o Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 o Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 o Neither the name of the orginal author Dale V. Patterson nor the names of any
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

Most of these constants are not used but are left as is for possible future use.

"""

__name__ = 'sockios_h'
__license__ = 'GPLv3'
__version__ = '0.0.2'
__date__ = 'February 2016'
__author__ = 'Dale Patterson'
__maintainer__ = 'Dale Patterson'
__email__ = 'wraith.wireless@yandex.com'
__status__ = 'Production'

# Routing table calls
SIOCADDRT = 0x890B # add routing table entry
SIOCDELRT = 0x890C # delete routing table entry
SIOCRTMSG = 0x890D # call to routing system

# Socket configuration controls from wireless.h
SIOCGIWNAME  = 0x8B01 # get name (standards can be retrieved here)
SIOCGIWTXPOW = 0x8B27 # get transmit power
SIOCGIWFREQ	 = 0x8B05 # get frequency
SIOCSIWSENS	 = 0x8B08 # set sensitivity (dBm) -> not supported
SIOCGIWSENS	 = 0x8B09 # get sensitivity (dBm) -> not supported
SIOCSIWTXPOW = 0x8B26 # set transmit power (dBm)

# Socket configuration controls
SIOCGIFNAME        = 0x8910 # get iface name
#SIOCSIFLINK        = 0x8911 # set iface channel
#SIOCGIFCONF        = 0x8912 # get iface list
SIOCGIFFLAGS       = 0x8913 # get flags
SIOCSIFFLAGS       = 0x8914 # set flags
SIOCGIFADDR        = 0x8915 # get PA address
SIOCSIFADDR        = 0x8916 # set PA address
#SIOCGIFDSTADDR     = 0x8917 # get remote PA address
#SIOCSIFDSTADDR     = 0x8918 # set remote PA address
SIOCGIFBRDADDR     = 0x8919 # get broadcast PA address
SIOCSIFBRDADDR     = 0x891a # set broadcast PA address
SIOCGIFNETMASK     = 0x891b # get network PA mask
SIOCSIFNETMASK     = 0x891c # set network PA mask
#SIOCGIFMETRIC      = 0x891d # get metric
#SIOCSIFMETRIC      = 0x891e # set metric
#SIOCGIFMEM         = 0x891f # get memory address (BSD)
#SIOCSIFMEM         = 0x8920 # set memory address (BSD)
#SIOCGIFMTU         = 0x8921 # get MTU size
#SIOCSIFMTU         = 0x8922 # set MTU size
#SIOCSIFNAME        = 0x8923 # set interface name
SIOCSIFHWADDR      = 0x8924 # set hardware address
#SIOCGIFENCAP       = 0x8925 # get/set encapsulations
#SIOCSIFENCAP       = 0x8926
SIOCGIFHWADDR      = 0x8927 # Get hardware address
#SIOCGIFSLAVE       = 0x8929 # Driver slaving support
#SIOCSIFSLAVE       = 0x8930
#SIOCADDMULTI       = 0x8931 # Multicast address lists
#SIOCDELMULTI       = 0x8932
SIOCGIFINDEX       = 0x8933 # name -> if_index mapping
SIOGIFINDEX  = SIOCGIFINDEX # misprint compatibility :-)
#SIOCSIFPFLAGS      = 0x8934 # set/get extended flags set
#SIOCGIFPFLAGS      = 0x8935
#SIOCDIFADDR        = 0x8936 # delete PA address
SIOCSIFHWBROADCAST = 0x8937 # set hardware broadcast addr
#SIOCGIFCOUNT       = 0x8938 # get number of devices
#SIOCGIFBR          = 0x8940 # Bridging support
#SIOCSIFBR          = 0x8941 # Set bridging options
#SIOCGIFTXQLEN      = 0x8942 # Get the tx queue length
#SIOCSIFTXQLEN      = 0x8943 # Set the tx queue length
#SIOCETHTOOL        = 0x8946 # Ethtool interface
#SIOCGMIIPHY        = 0x8947 # Get address of MII PHY in use
#SIOCGMIIREG        = 0x8948 # Read MII PHY register.
#SIOCSMIIREG        = 0x8949 # Write MII PHY register.
#SIOCWANDEV         = 0x894A # get/set netdev parameters
#SIOCOUTQNSD        = 0x894B # output queue size (not sent only)

# ARP cache control calls
#SIOCDARP = 0x8953 # delete ARP table entry
#SIOCGARP = 0x8954 # get ARP table entry
#SIOCSARP = 0x8955 # set ARP table entry

# RARP cache control calls
#SIOCDRARP = 0x8960 # delete RARP table entry
#SIOCGRARP = 0x8961 # get RARP table entry
#SIOCSRARP = 0x8962 # set RARP table entry

# Driver configuration calls
#SIOCGIFMAP = 0x8970 # Get device parameters
#SIOCSIFMAP = 0x8971 # Set device parameters

# DLCI configuration calls
#SIOCADDDLCI	= 0x8980 # Create new DLCI device
#SIOCDELDLCI	= 0x8981 # Delete DLCI device

#SIOCGIFVLAN	= 0x8982 # 802.1Q VLAN support
#SIOCSIFVLAN	= 0x8983 # Set 802.1Q VLAN options

# bonding calls
#SIOCBONDENSLAVE	       = 0x8990	# enslave a device to the bond
#SIOCBONDRELEASE        = 0x8991	# release a slave from the bond
#SIOCBONDSETHWADDR      = 0x8992	# set the hw addr of the bond
#SIOCBONDSLAVEINFOQUERY = 0x8993 # rtn info about slave state
#SIOCBONDINFOQUERY      = 0x8994	# rtn info about bond state
#SIOCBONDCHANGEACTIVE   = 0x8995 # update to a new active slave

# bridge calls
#SIOCBRADDBR = 0x89a0 # create new bridge device
#SIOCBRDELBR = 0x89a1 # remove bridge device
#SIOCBRADDIF = 0x89a2 # add interface to bridge
#SIOCBRDELIF = 0x89a3 # remove interface from bridge

# hardware time stamping: parameters in linux/net_tstamp.h
#SIOCSHWTSTAMP = 0x89b0 # set and get config
#SIOCGHWTSTAMP = 0x89b1 # get config


#!/usr/bin/env python
""" pyric Python Radio Interface Controller

Copyright (C) 2016  Dale V. Patterson (wraith.wireless@yandex.com)

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

Redistribution and use in source and binary forms, with or without modifications,
are permitted provided that the following conditions are met:
 o Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 o Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 o Neither the name of the orginal author Dale V. Patterson nor the names of any
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

Defines the PyRIC error class and constants for some errors. All pyric errors
will follow the 2-tuple form of EnvironmentError. Also defines constansts for
PyPI packaging.

WARNING: DO NOT import *

Requires:
 linux (3.x or 4.x kernel)
 Python 2.7

 pyric 0.1.5 through 0.1.6
  desc: wireless nic library: wireless radio identification, manipulation, enumeration
   concentrate on STA/AP related functionality
  includes: /nlhelp /lib /net /utils pyw.py
  changes:
   See CHANGES in top-level directory
"""

__name__ = 'pyric'
__license__ = 'GPLv3'
__version__ = '0.1.6.4'
__date__ = 'January 2017'
__author__ = 'Dale Patterson'
__maintainer__ = 'Dale Patterson'
__email__ = 'wraith.wireless@yandex.com'
__status__ = 'Production'

"""
 define pyric exceptions
  all exceptions are tuples t=(error code,error message)
  we use error codes defined in errno, using EUNDEF = -1 to define an undefined
  error I don't like importing all from errno but it provides conformity in
  error handling i.e modules using pyric.error do not need to call pyric.EUNDEF
  and errno.EINVAL but can call pyric.EUNDEF and pyric.EINVAL
"""
EUNDEF = -1                   # undefined error
from errno import *           # make all errno errors pyric errors
errorcode[EUNDEF] = "EUNDEF"  # add ours to errorcode dicts
class error(EnvironmentError):
    def __init__(self,errno,errmsg=None):
        if not errmsg: errmsg = strerror(errno)
        EnvironmentError.__init__(self,errno,errmsg)

def strerror(errno):
    """
    :param errno: error code
    :returns: error message
    """
    import os
    if errno < 0: return "Undefined error"
    elif errno == EPERM: return "Superuser privileges required"
    elif errno == EINVAL: return "Invalid parameter"
    elif errno == EBUSY:
        msg = "{0}. Make sure Card is up and no other devices share the same phy"
        return msg.format(os.strerror(EBUSY))
    elif errno == ENFILE: return "There are no available netlink sockets"
    else:
        return os.strerror(errno)

# for setup.py use
# redefine version for easier access
version = __version__

# define long description
long_desc = """
# PyRIC 0.1.6.4: Python Radio Interface Controller
## Linux wireless library for the Python Wireless Developer and Pentester

## DESCRIPTION:
PyRIC (is a Linux only) library providing wireless developers and pentesters the
ability to identify, enumerate and manipulate their system's wireless cards
programmatically in Python. Pentesting applications and scripts written in Python
have increased dramatically in recent years. However, these tools still rely on
Linux command lines tools for setup/preparation and restoration of the system for
use. Until now. Why use subprocess.Popen, regular expressions and str.find to
interact with your wireless cards? PyRIC is:

1. Pythonic: no ctypes, SWIG etc. PyRIC redefines C header files as Python and
uses sockets to communicate with the kernel.
2. Self-sufficient: No third-party files used. PyRIC is completely self-contained.
3. Fast: (relatively speaking) PyRIC is faster than using command line tools
through subprocess.Popen
4. Parseless: Get the output you want without parsing output from command line
tools. Never worry about newer iw versions and having to rewrite your parsers.
5. Easy: If you can use iw, you can use PyRIC.

## CURRENT STATE
ATT, PyRIC pyw provides the following:
* enumerate interfaces and wireless interfaces
* identify a cards chipset and driver
* get/set hardware address
* get/set ip4 address, netmask and or broadcast
* turn card on/off
* get supported standards
* get supported commands
* get supported modes
* get dev info
* get phy info
* get link info
* get/set regulatory domain
* get/set mode
* get/set coverage class, RTS threshold, Fragmentation threshold & retry limits
* add/delete interfaces
* enumerate ISM and UNII channels
* block/unblock rfkill devices
* check 'connectivity', disconnect from AP

In utils, several helpers can be found that can be used to:
* enumerate channels and frequencies and convert between the two
* manipulate mac addresses and generate random ones
* fetch and parse the IEEE oui text file
* further rfkill operations to include listing all rfkill devices

At it's heart, PyRIC is a Python port of (a subset of) iw and by extension, a
Python port of Netlink w.r.t nl80211 functionality. The original goal of PyRIC
was to provide a simple interface to the underlying nl80211 kernel support,
handling the complex operations of Netlink seamlessy while maintaining a minimum
of "code walking" to understand, modify and extend. But, why stop there? Since
it's initial inception, PyRIC has grown. PyRIC puts iw, ifconfig, rfkill, udevadm,
airmon-ng and macchanger in your hands (or your program).
"""



#!/usr/bin/env python

""" ouifetch.py: IEEE oui helper

Copyright (C) 2016  Dale V. Patterson (wraith.wireless@yandex.com)

This program is free software:you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation,either version 3 of the License,or (at your option) any later
version.

Redistribution and use in source and binary forms,with or without modifications,
are permitted provided that the following conditions are met:
 o Redistributions of source code must retain the above copyright notice,this
   list of conditions and the following disclaimer.
 o Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 o Neither the name of the orginal author Dale V. Patterson nor the names of any
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

oui dict as determined by IEEE oui.txt

"""
from __future__ import print_function  # python 2to3 compability

__name__ = 'ouifetch'
__license__ = 'GPLv3'
__version__ = '0.0.2'
__date__ = 'July 2016'
__author__ = 'Dale Patterson'
__maintainer__ = 'Dale Patterson'
__email__ = 'wraith.wireless@yandex.com'
__status__ = 'Production'

try:
    # load urllib related for python 2
    # noinspection PyCompatibility
    from urllib2 import Request as url_request
    # noinspection PyCompatibility
    from urllib2 import urlopen as url_open
    # noinspection PyCompatibility
    from urllib2 import URLError as url_error
except ImportError:
    # noinspection PyCompatibility,PyUnresolvedReferences
    from urllib.request import Request as url_request
    # noinspection PyCompatibility,PyUnresolvedReferences
    from urllib.request import urlopen as url_open
    # noinspection PyUnresolvedReferences
    from urllib import error as url_error
import os,sys,datetime,time
import pyric

OUIURL = 'http://standards-oui.ieee.org/oui.txt'
OUIPATH = os.path.join(os.path.dirname(os.path.abspath(__file__)),'data/oui.txt')

def load(opath=None):
    """
     parse oui.txt file
     :param opath: path of oui text file
     :returns: oui dict {oui:manuf} for each oui in path or empty dict
    """
    fin = None
    ouis = {}

    if not opath: opath = OUIPATH

    try:
        fin = open(opath)
        for line in fin.readlines()[1:]:
            o,m = line.strip().split('\t')
            ouis[o.lower()] = m[0:100]
        fin.close()
    except IndexError:
        pass
    except IOError as e:
        raise pyric.error(e.errno,e.strerror)
    finally:
        if fin and not fin.closed: fin.close()
    return ouis

def fetch(opath=None,verbose=False):
    """
     retrieves oui.txt from IEEE and writes to data file
     :param opath: fullpath of oui.txt
     :param verbose: write updates to stdout
    """
    # determine if data path is legit
    if opath is None: opath = OUIPATH
    if not os.path.isdir(os.path.dirname(opath)):
        print("Path to data is incorrect {0}".format(opath))
        sys.exit(1)

    # fetch oui file from ieee
    fout = None

    # set up url request
    req = url_request(OUIURL)
    req.add_header('User-Agent',"PyRIC +https://github.com/wraith-wireless/PyRIC/")
    try:
        # retrieve the oui file and parse out generated date
        if verbose: print('Fetching ', OUIURL)
        res = url_open(req)
        if verbose: print("Parsing OUI file")

        if verbose: print("Opening data file {0} for writing".format(opath))
        fout = open(opath,'w')
        gen = datetime.datetime.utcnow().isoformat() # use current time as the first line
        fout.write(gen+'\n')

        # pull out ouis
        t = time.time()
        cnt = 0
        for l in res:
            if '(hex)' in l:
                # extract oui and manufacturer
                oui,manuf = l.split('(hex)')
                oui = oui.strip().replace('-',':')
                manuf = manuf.strip()
                if manuf.startswith("IEEE REGISTRATION AUTHORITY"):
                    manuf = "IEEE REGISTRATION AUTHORITY"

                # write to file & update count
                fout.write('{0}\t{1}\n'.format(oui,manuf))
                cnt += 1
                if verbose: print("{0}:\t{1}\t{2}".format(cnt,oui,manuf))
        print("Wrote {0} OUIs in {1:.3} secs".format(cnt,time.time()-t))
    except url_error as e:
        print("Error fetching oui file: {0}".format(e))
    except IOError as e:
        print("Error opening output file {0}".format(e))
    except Exception as e:
        print("Error parsing oui file: {0}".format(e))
    finally:
        if fout: fout.close()

#if __name__ == '__main__':
#    # create arg parser and parse command line args
#    print "OUI Fetch {0}".format(__version__)
#    argp = ap.ArgumentParser(description="IEEE OUI fetch and parse")
#    argp.add_argument('-p','--path',help="Path to write parsed file")
#    argp.add_argument('-v','--verbose',action='store_true',help="Display operations to stdout")
#    argp.add_argument('--version',action='version',version="OUI Fetch {0}".format(__version__))
#    args = argp.parse_args()
#    verbose = args.verbose
#    path = args.path

    # execute
#    fetch(path,verbose)



#!/usr/bin/env python

""" channels.py:802.11 channel/freq utilities

Copyright (C) 2016  Dale V. Patterson (wraith.wireless@yandex.com)

This program is free software:you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation,either version 3 of the License,or (at your option) any later
version.

Redistribution and use in source and binary forms,with or without modifications,
are permitted provided that the following conditions are met:
 o Redistributions of source code must retain the above copyright notice,this
   list of conditions and the following disclaimer.
 o Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 o Neither the name of the orginal author Dale V. Patterson nor the names of any
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

Defines ISM 2.4Ghz, UNII 5Ghz and 4.9GHz frequencies and channels

Need to define 3GHz channels/freqs

"""

__name__ = 'channels'
__license__ = 'GPLv3'
__version__ = '0.0.2'
__date__ = 'August 2016'
__author__ = 'Dale Patterson'
__maintainer__ = 'Dale Patterson'
__email__ = 'wraith.wireless@yandex.com'
__status__ = 'Production'

import pyric.net.wireless.nl80211_h as nl80211h

# redefined widths (allowed in nl80211h)
CHTYPES = nl80211h.NL80211_CHAN_TYPES   # widths nl80211 supports i.e HT40- 
CHWIDTHS = nl80211h.NL80211_CHAN_WIDTHS # actual widths

# ISM Bands (ieee80211.h-> BAND_ID_2G)
ISM_24_C2F={1:2412,2:2417,3:2422,4:2427,5:2432,6:2437,7:2442,
            8:2447,9:2452,10:2457,11:2462,12:2467,13:2472,14:2484}
ISM_24_F2C={2432:5,2467:12,2437:6,2472:13,2442:7,2484:14,2412:1,
            2447:8,2417:2,2452:9,2422:3,2457:10,2427:4,2462:11}

# UNII 5 Bands (ieee80211.h-> BAND_ID_5G)
# confirm that ch 34, ch 54, 58, 62, 102, 106, 110, 114 118, 122, 126, 138, 144, 151,
# 155, 159,
UNII_5_C2F={34:5170,36:5180,38:5190,40:5200,42:5210,44:5220,46:5230,48:5240,50:5250,
            52:5260,54:5270,56:5280,58:5290,60:5300,62:5310,64:5320,100:5500,
            102:5510,104:5520,106:5530,108:5540,110:5550,112:5560,114:5570,116:5580,
            118:5590,120:5600,122:5610,124:5620,126:5630,128:5640,132:5660,136:5680,
            138:5690,140:5700,142:5710,144:5720,149:5745,151:5755,153:5765,155:5775,
            157:5785,159:5795,161:5805,165:5825}
UNII_5_F2C={5250:50,5765:153,5510:102,5640:128,5260:52,5775:155,5520:104,5270:54,
            5785:157,5530:106,5660:132,5280:56,5795:159,5540:108,5290:58,5805:161,
            5550:110,5680:136,5170:34,5300:60,5560:112,5690:138,5180:36,5310:62,
            5825:165,5570:114,5700:140,5190:38,5320:64,5580:116,5710:142,5200:40,
            5590:118,5720:144,5210:42,5600:120,5220:44,5610:122,5230:46,5745:149,
            5620:124,5240:48,5755:151,5500:100,5630:126}

# UNII 4 Bands (ieee80211.h-> BAND_ID_5G)
UNII_4_C2F={183:4915,184:4920,185:4925,187:4935,188:4940,189:4945,192:4960,196:4980}
UNII_4_F2C={4960:192,4935:187,4940:188,4945:189,4915:183,4980:196,4920:184,4925:185}

# US high powered backhaul (ieee80211.h-> BAND_ID_3G)
#131 	3657.5 	 132 	36622.5 ? 132 	3660.0 133 	3667.5 133 	3665.0
#134 	3672.5 	 134 	3670.0    135 	3677.5 136 	3682.5 136 	3680.0
#137 	3687.5 	 137 	3685.0    138 	3689.5 138 	3690.0

def channels(band=None):
    """
     returns list of channels
     :param band: one of {None=all|'ISM'=2.4GHz|'UNII'=4.9/5GHz|'UNII5'=5GHz,
      'UNII4'=4GHz}
     :returns:list of channels
    """
    if band == 'ISM': return ISM_24_C2F.keys()
    elif band == 'UNII': return UNII_5_C2F.keys() + UNII_4_C2F.keys()
    elif band == 'UNII4': return UNII_4_C2F.keys()
    elif band == 'UNII5': return UNII_5_C2F.keys()
    try:
        return sorted(ISM_24_C2F.keys() + UNII_5_C2F.keys() + UNII_4_C2F.keys())
    except TypeError:
        # python 3 doesn't like the above (uses dict_keys obj instead of list)
        return sorted(list(ISM_24_C2F.keys()) + list(UNII_5_C2F.keys()) + list(UNII_4_C2F.keys()))

def freqs(band=None):
    """
     returns list of channels
     :param band: one of {None=all|'ISM'=2.4GHz|'UNII'=4.9/5GHz|'UNII5'=5GHz,
      'UNII4'=4GHz}
     :returns:list of frequencies
    """
    if band == 'ISM': return sorted(ISM_24_F2C.keys())
    elif band == 'UNII': return sorted(UNII_5_F2C.keys() + UNII_4_F2C.keys())
    elif band == 'UNII4': return sorted(UNII_4_F2C.keys())
    elif band == 'UNII5': return sorted(UNII_5_F2C.keys())
    return sorted(ISM_24_F2C.keys() + UNII_5_F2C.keys()+ UNII_4_F2C.keys())

def ch2rf(c):
    """
     channel to frequency conversion
     :param c:channel
     :returns:frequency in MHz corresponding to channel
    """
    if c in ISM_24_C2F: return ISM_24_C2F[c]
    if c in UNII_5_C2F: return UNII_5_C2F[c]
    if c in UNII_4_C2F: return UNII_4_C2F[c]
    return None

def rf2ch(f):
    """
     frequency to channel conversion
     :param f:frequency (in MHz)
     :returns:channel number corresponding to frequency
    """
    if f in ISM_24_F2C: return ISM_24_F2C[f]
    if f in UNII_5_F2C: return UNII_5_F2C[f]
    if f in UNII_4_F2C: return UNII_4_F2C[f]
    return None



#!/usr/bin/env python
""" libnl provides libnl(ish) functionality

Copyright (C) 2016  Dale V. Patterson (wraith.wireless@yandex.com)

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

Redistribution and use in source and binary forms, with or without modifications,
are permitted provided that the following conditions are met:
 o Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 o Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 o Neither the name of the orginal author Dale V. Patterson nor the names of any
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

Relates similarily to libnl by providing functions handling netlink messages
and sockets. Where possible I have attempted to name the below functions the
same as would be found in libnl to ease any transitions. However, I have taken
liberties with the below as these functions only handle nl80211 generic netlink
messages.

Provides access to netlink sockets and messages in a manner similar to libnl.
see http://www.carisma.slowglass.com/~tgr/libnl/doc/core.html

"""

__name__ = 'libnl'
__license__ = 'GPLv3'
__version__ = '0.1.1'
__date__ = 'July 2016'
__author__ = 'Dale Patterson'
__maintainer__ = 'Dale Patterson'
__email__ = 'wraith.wireless@yandex.com'
__status__ = 'Production'

from time import time
from os import getpid,strerror
import struct
import socket
from binascii import hexlify
import errno
import pyric.net.netlink_h as nlh
import pyric.net.genetlink_h as genlh
from pyric.net.policy import nla_datatype
import sys
_PY3_ = sys.version_info.major == 3

class error(EnvironmentError): pass

BUFSZ = 32768 # Generic default buffersize

"""
NETLINK SOCKET
"""

class NLSocket(dict):
    """
     Wrapper around a Netlink socket. Exposes the following properties: (callable
     by '.')
      sock: socket, get only
      fd: file descriptor, get only
      pid: (local) port id, get/set
      grpm: group mask, get/set
      seq: seq. #, get/set
      tx: tx buffer size, get only
      rx: rx buffer size, get only
    """
    def __new__(cls,d=None):
        return super(NLSocket,cls).__new__(cls,dict({} if not d else d))

    def __repr__(self):
        """ :returns: description """
        fmt = "NLSocket(fd: {0}, pid: {1}, grpm: {2}, seq: {3}, tx: {4}, rx: {5})"
        return fmt.format(self.fd,self.pid,self.grpm,self.seq,self.tx,self.rx)

    @property
    def sock(self): return self['sock']

    @property
    def fd(self): return self['sock'].fileno()

    @property
    def tx(self): return self['sock'].getsockopt(socket.SOL_SOCKET,socket.SO_SNDBUF)

    @tx.setter
    def tx(self,v):
        if v < 128 or v > _maxbufsz_():
            raise error(errno.EINVAL,"Invalid buffer size")
        self['sock'].setsockopt(socket.SOL_SOCKET,socket.SO_SNDBUF,v)

    @property
    def rx(self): return self['sock'].getsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF)

    @rx.setter
    def rx(self,v):
        if v < 128 or v > _maxbufsz_():
            raise error(errno.EINVAL,"Invalid buffer size")
        self['sock'].setsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF,v)

    @property
    def pid(self): return self['pid']

    @pid.setter
    def pid(self,v):
        if v < 1: raise error(errno.EINVAL,"Invalid port id")
        self['pid'] = v

    @property
    def grpm(self): return self['grpm']

    @grpm.setter
    def grpm(self,v): self['grpm'] = v

    @property
    def seq(self): return self['seq']

    @seq.setter
    def seq(self,v):
        if v < 1: raise error(errno.EINVAL,"Invalid sequence number")
        self['seq'] = v

    @property
    def timeout(self): return self['sock'].gettimeout()

    @timeout.setter
    def timeout(self,v):
        if v and v < 0: raise error(errno.EINVAL,"Invalid timeout value")
        self['sock'].settimeout(v)

    #### wrap socket functions

    def incr(self):
        """ increments seq num """
        self['seq'] += 1

    def send(self,pkt):
        """
         send data
         :param pkt: data to be sent
         :returns: bytes sent
        """
        return self['sock'].send(pkt)

    def recv(self):
        """ :returns: msg from kernel """
        return self['sock'].recv(self.rx)

    def close(self):
        """ closes the socket """
        if self['sock']: self['sock'].close()
        self['sock'] = self['fid'] = None
        self['pid'] = self['grpm'] = self['seq'] = None
        self['rx'] = self['tx'] = None

def nl_socket_alloc(pid=None,grps=0,seq=None,rx=None,tx=None,timeout=None):
    """
     create a netlink socket
     :param pid: port id
     :param grps: multicast groups mask
     :param seq: initial seq. #
     :param rx: rx buffer size
     :param tx: tx buffer size
     :param timeout: time to block on socket
     :returns: a NLSocket
     NOTE:
      per man socket, the kernel will double the specified rx/tx buffer size and
      the min. size is 128
    """
    # set & validate paramaters
    pid = pid or getpid() + int(time()) # allow multiple sockets on this host
    if pid < 1: raise error(errno.EINVAL,"Invalid port id")
    seq = seq or int(time())
    if seq < 1: raise error(errno.EINVAL,"Invalid sequence number")
    rx = rx or BUFSZ
    if rx < 128 or rx > _maxbufsz_(): raise error(errno.EINVAL,"Invalid rx size")
    tx = tx or BUFSZ
    if tx < 128 or tx > _maxbufsz_(): raise error(errno.EINVAL,"Invalid tx size")

    # create the socket and return it
    try:
        s = socket.socket(socket.AF_NETLINK,socket.SOCK_RAW,nlh.NETLINK_GENERIC)
        s.setsockopt(socket.SOL_SOCKET,socket.SO_SNDBUF,tx)
        s.setsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF,rx)
        s.settimeout(timeout)
        s.bind((pid,grps))
    except socket.error as e:
        raise error(e.errno,e.strerror)
    return NLSocket({'sock':s,'tx':tx,'rx':rx,'pid':pid,'grpm':grps,'seq':seq})

def nl_socket_free(sock):
    """ close the socket """
    try:
        sock.close()
    except AttributeError: # sock already closed
        pass
    return None

def nl_socket_pid(sock):
    """ :returns netlink socket's pid """
    return sock.pid

def nl_socket_grpmask(sock):
    """ :returns netlink socket's group mask """
    return sock.grpm

def nl_sendmsg(sock,msg,override=False):
    """
      sends msg to kernel
      :param sock: netlink socket
      :param msg: nlmsg stream
      :param override: if True will keep the message's pid and seq. This can
      be used for testing
    """
    try:
        # change the msg's pid & seq to that of the sockets prior to sending &
        # set the ack flag - I can't figure how to tell in recv if an ack was
        # requested or not so I force an ACK here
        if not override:
            msg.pid = sock.pid
            msg.seq = sock.seq
        msg.flags = msg.flags | nlh.NLM_F_ACK
        sent = sock.send(msg.tostream())
        if sent != msg.len: raise error(errno.EBADMSG,"Message sent incomplete")
    except socket.error as e:
        raise error(errno.ECOMM, e)
    except AttributeError:
        raise error(errno.ENOTSOCK,"Invalid netlink socket")

def nl_recvmsg(sock):
    """
     :param sock: netlink socket
     :returns: a GENLMsg received from the socket
    """
    try:
        # we can expect two types of messages 1) an instant success message
        # or 2) a data message. If it's a data message, we need to catch
        # the ack
        msg = nlmsg_fromstream(sock.recv())
        try:
            # catch the follow on ack
            _ = nlmsg_fromstream(sock.recv())
        except error as e:
            # on success, just return the orginal message
            if e.errno == nlh.NLE_SUCCESS: pass
            else: raise
        if sock.seq != msg.seq: raise error(errno.EBADMSG,"Seq. # out of order")
        return msg
    except socket.timeout:
        raise error(-1,"Socket timed out")
    #except socket.error as e: # this became in issue in python 3
    #    raise error(errno.ENOTSOCK,e)
    except error as e:
        if e.errno == nlh.NLE_SUCCESS: return nlh.NLE_SUCCESS
        raise # rethrow
    finally:
        # always increment the sequence #
        sock.incr()

"""
NETLINK MESSAGES

generic netlink data exchanged between user and kernel space is a netlink message
of type NETLINK_GENERIC using the netlink attributes interface. Messages are in
the format:

  <------- NLMSG_ALIGN(hlen) ------> <---- NLMSG_ALIGN(len) --->
 +----------------------------+-----+---------------------+-----+
 |           Header           | Pad |       Payload       | Pad |
 |      struct nlmsghdr       |     |                     |     |
 +----------------------------+-----+---------------------+-----+

  <-------- GENL_HDRLEN -------> <--- hdrlen -->
                                 <------- genlmsg_len(ghdr) ------>
 +------------------------+-----+---------------+-----+------------+
 | Generic Netlink Header | Pad | Family Header | Pad | Attributes |
 |    struct genlmsghdr   |     |               |     |            |
 +------------------------+-----+---------------+-----+------------+

 <-------- nla_attr_size(payload) --------->
 +------------------+-----+------------------+-----+
 | Attribute Header | Pad |     Payload      | Pad |
 +------------------+-----+------------------+-----+

Example: nlmsg for retrieving the family id of nl80211
 |<----------- nlmsghdr ---------->|gemsghdr|<-------- attr -------->|
 |                                          |< hdr >|<--- payload -->|
 | 0 1 2 3 4 5 6 7| 0 1 2 3 4 5 6 7| 0 1 2 3 4 5 6 7| 0 1 2 3 4 5 6 7|
 |2000000010000500|02000000626b0000|030100000c000200|6e6c383032313100|

Netlink message components are aligned on boundaries of 4
"""

"""
The GENLMsg class. There are two methods of creating a GENLMsg:
 1) nlmsg_new -> create a new 'default' msg
 2) nlmsg_fromstream -> create a msg from a string

NOTE: regardles of the message's specified port id & seq #, when sending they
 will be set to that of the socket's port id & seq #. The can be set during
 creation IOT facilitate testing etc
"""

class GENLMsg(dict):
    """
     A wrapper around dict for an underlying generic netlink message of nl80211
     family. Exposes the following properties: (callable by '.')
      len: message length, get only
      nltype: netlink type, get/set
      flags: message flags, get/set
      seq: seq. #, get/set
      pid: port id, get/set
      cmd: etlink command, get/set
      attrs: message attributes get only
     Each attributes is a tuple t = (attribute,value,datatype)
      attribute: netlink type of attribute like CTRL_ATTR_FAMILY_ID
      value: actual value (i.e. unpacked)
      datatype: datatype of attribute value as defined in netlink_h i.e. NLA_U8
     and are sent in the order they are put on the attr list
    """
    def __new__(cls,d=None):
        return super(GENLMsg,cls).__new__(cls,dict({} if not d else d))

    def __repr__(self):
        fmt = "nlmsghdr(len={0},type={1},flags={2},seq={3},pid={4})\n"
        ret = fmt.format(self.len,self.nltype,self.flags,self.seq,self.pid)
        ret += "genlmsghdr(cmd={0})\n".format(self.cmd)
        ret += "attributes:\n"
        for i,(a,v,d) in enumerate(self.attrs):
            # getting character(s) in some bytestrings that cause the terminal to
            # hang (why?) hexlify unspec and unpacked nested to avoid this
            if d == nlh.NLA_UNSPEC:
                v = hexlify(v)
            elif d == nlh.NLA_NESTED:
                v = [(idx,hexlify(attr)) for idx,attr in v]
            ret += "\t{0}: type={1},datatype={2}\n\tvalue={3}\n".format(i,a,d,v)
        return ret

    #### PROPERTIES

    @property # length (inlcuding padding and headers)
    def len(self): return len(self.tostream())

    @property
    def vers(self): return 1

    @property
    def nltype(self): return self['type']

    @nltype.setter
    def nltype(self,v):
        if v < 0: raise error(errno.ERANGE,"Netlink type {0} is invalid".format(v))
        self['type'] = v

    @property
    def flags(self): return self['flags']

    @flags.setter
    def flags(self,v): self['flags'] = v

    @property
    def seq(self): return self['seq']

    @seq.setter
    def seq(self,v):
        if v < 1: raise error(errno.ERANGE,"Invalid seq. number")
        self['seq'] = v

    @property
    def pid(self): return self['pid']

    @pid.setter
    def pid(self,v):
        if v < 1: raise error(errno.ERANGE,"Invalid port id")
        self['pid'] = v

    @property
    def cmd(self): return self['cmd']

    @cmd.setter
    def cmd(self,v):
        if v < 0: raise error(errno.ERANGE,"Invalid cmd")
        self['cmd'] = v

    @property
    def attrs(self): return self['attrs']

    @property
    def numattrs(self): return len(self['attrs'])

    #### METHODS

    def tostream(self):
        """ :returns packed netlink message """
        payload = genlh.genlmsghdr(self['cmd'])  # nlhsghdr, genlmsghdr end at boundary of 4
        for a,v,d in self['attrs']:
            try:
                payload += _attrpack_(a,v,d)
            except (TypeError,AttributeError,struct.error) as e:
                #if d == nlh.NLA_NESTED: pass # we need to fix here
                raise error(-1,"Packing {0} {1}: {2}".format(a,v,e))
        return nlh.nlmsghdr(len(payload),self.nltype,self.flags,self.seq,self.pid) + payload

def nlmsg_new(nltype=None,cmd=None,seq=None,pid=None,flags=None,attrs=None):
    """
     :param nltype: message content
     :param cmd: genetlink service type
     :param seq: sequence number
     :param pid: port id
     :param flags: additional flags
     :param attrs: attr list list of tuples t = (attribute,value,attr_datatype)
      attribute = netlinke type of attribute like CTRL_ATTR_FAMILY_ID
      value = actual value (i.e. unpacked)
      attr_datatype = type of attribute value as defined in netlink_h i.e. NLA_U8
     :returns a GENLMsg
     NOTE:
      # version is hardcoded as 1 and len is calculated
    """
    return GENLMsg({'type':nltype or nlh.NETLINK_GENERIC,
                    'flags':flags or (nlh.NLM_F_REQUEST|nlh.NLM_F_ACK),
                    'seq':seq or int(time()),
                    'pid':pid or getpid(),
                    'cmd':cmd or genlh.CTRL_CMD_UNSPEC,
                    'attrs':attrs or []})

def nlmsg_fromstream(stream,override=False):
    """
     create a GENLMsg from a stream
     :param stream: packed binary data
     :param override: override ack processings - DO NOT USE for debugging only
     :returns: a GENLMsg
    """
    # parse out netlink/generic netlink headers
    try:
        l,t,fs,s,p = struct.unpack_from(nlh.nl_nlmsghdr,stream,0)
        if t == nlh.NLMSG_ERROR or (l == nlh.NLMSGACKLEN and not override):
            # have an (possible) ack/nack i.e. error msg
            e = struct.unpack_from(nlh.nl_nlmsgerr,stream,nlh.NLMSGHDRLEN)[0]
            raise error(abs(e),strerror(abs(e)))
        c,_,_ = struct.unpack_from(genlh.genl_genlmsghdr,stream,nlh.NLMSGHDRLEN)
    except struct.error as e:
        raise error(-1,"Error parsing headers: {0}".format(e))

    # create a new message with hdr values then parse the attributes
    msg = nlmsg_new(t,c,s,p,fs)
    nla_parse(msg,l,t,stream,nlh.NLMSGHDRLEN + genlh.GENLMSGHDRLEN)
    return msg

def nla_parse(msg,l,mtype,stream,idx):
    """
     parses attributes in stream, putting them in msg
     :param msg: current message
     :param l: total length of message
     :param mtype: message content
     :param stream: byte stream
     :param idx: current index in stream
    """
    # get policy family NOTE: cheating here, we know it's either generic or nl80211
    pol = 'ctrl_attr' if mtype == nlh.NETLINK_GENERIC else 'nl80211_attr'
    attrlen = nlh.NLATTRHDRLEN # pull out these to avoid
    attrhdr = nlh.nl_nlattrhdr # doing so in each iteration

    # eat the stream until the end
    while idx < l:
        a = atype = alen = None  # shut pycharm up about unitialized variable
        try:
            alen,atype = struct.unpack_from(attrhdr,stream,idx)  # get length, type
            idx += attrlen               # move to attr start
            alen -= attrlen              # attr length (w/ padding)
            a = stream[idx:idx+alen]     # attr value
            dt = nla_datatype(pol,atype) # attr datatype

            # Note: we use unpack_from which will ignore the null bytes in numeric
            # datatypes & for strings, strip trailing null bytes
            # dt == nlh.NLA_UNSPEC: ignore
            if _PY3_ and (dt == nlh.NLA_STRING or dt == nlh.NLA_UNSPEC):
                # python 3 returns a bytes object, convert to string
                try:
                    a = a.decode('ascii')
                except UnicodeDecodeError:
                    pass # Fuck You Python 3
            if dt == nlh.NLA_STRING: a = _nla_strip_(a)
            elif dt == nlh.NLA_U8: a = struct.unpack_from("B",a,0)[0]
            elif dt == nlh.NLA_U16: a = struct.unpack_from("H",a,0)[0]
            elif dt == nlh.NLA_U32: a = struct.unpack_from("I",a,0)[0]
            elif dt == nlh.NLA_U64: a = struct.unpack_from("Q",a,0)[0]
            elif dt == nlh.NLA_FLAG: a = ''  # flags should be 0 size
            elif dt == nlh.NLA_MSECS: a = struct.unpack_from("Q",a,0)[0]
            elif dt == nlh.NLA_SET_U8: a = nla_parse_set(a,nlh.NLA_U8)
            elif dt == nlh.NLA_SET_U16: a = nla_parse_set(a,nlh.NLA_U16)
            elif dt == nlh.NLA_SET_U32: a = nla_parse_set(a,nlh.NLA_U32)
            elif dt == nlh.NLA_SET_U64: a = nla_parse_set(a,nlh.NLA_U64)
            elif dt == nlh.NLA_NESTED: a = nla_parse_nested(a)
            nla_put(msg,a,atype,dt)
        except struct.error: # append as Error, stripping null bytes
            nla_put(msg,_nla_strip_(a),atype,nlh.NLA_ERROR)
        except error as e:
            if e.errno == errno.EINVAL: # a nested or set failed to parse correctly
                nla_put(msg, _nla_strip_(a), atype, nlh.NLA_ERROR)
            else:
                raise
        except MemoryError as e: # hopefully don't get here
            raise error(-1,"Attr type {0} of pol {1} failed: {2}".format(atype,pol,e))
        idx = nlh.NLMSG_ALIGN(idx + alen)  # move index to next attr

def nla_parse_nested(nested):
    """
     :param nested: the nested attribute with attribute header removed
     :returns: list of tuples (index, data) where index is the index
      into an enum structure and attribute is the packed attribute)
      Callers must parse these themselves - see below for details

      From libnl (Thomas Graf)
      When nesting attributes, the nested attributes are included as payload of
      a container attribute. Attributes are nested by surrounding them with calls
      to nla_nest_start() and nla_nest_end().

       <-------- nla_attr_size(payload) --------->
       +------------------+-----+------------------+-----+
       | Attribute Header | Pad |     Payload      | Pad |
       +------------------+-----+------------------+-----+

      Nested attributes (after removing the Attribute header) should have the
      form:

       +--------+--------+-----+
       | Length |Payload | Pad |
       +--------+--------+-----+
       <-- 2 --><- var ->< var >

      where:
      Length (u16) is the length of the nested attribute (excluding padding
       affixed to the end to align the attribute). The size of padding is
       determined by NLMSG_ALIGN
      Payload has the format:

      +-------+--------+-----+
      | Index |  Data  | Pad |
      +-------+--------+-----+
      <-- 2 --><- var ->< var >

      where index is the index into an enum structure as determined by the
      attribute type of the nested attribute which is found in the Attribute
      Header
    """
    ns = []
    idx =  0
    l = len(nested)
    while idx < l:
        # first two bytes define length, including these bytes, length does not
        # include pad byte(s) affixed to end for proper alignment
        alen = struct.unpack_from('H', nested, idx)[0]
        if alen == 0: raise error(errno.EINVAL, "Invalid nesting")
        #ns.append(nested[idx+2:idx+alen]) # don't include the length bytes
        nattr = nested[idx + 2:idx + alen]
        ns.append((struct.unpack_from('H',nattr,0)[0],nattr[2:]))
        idx += nlh.NLMSG_ALIGN(alen)
    return ns

def nla_parse_set(aset,etype):
    """
     parses out a set of like sized elements
     :param aset: a netlink nl80211 set
     :param etype: data type of each element in the set
     :returns: list of elements in aset
    """
    # get the struct format and element size
    if etype == nlh.NLA_U8: fmt = "B"
    elif etype == nlh.NLA_U16: fmt = "H"
    elif etype == nlh.NLA_U32: fmt = "I"
    elif etype == nlh.NLA_U64: fmt = "Q"
    else:
        raise error(errno.EINVAL,"Set elements are not valid datatype")
    esize = struct.calcsize(fmt)

    ss = []
    idx = 0
    asize = len(aset)
    while idx < asize:
        if asize - idx < esize: break # don't attempt to parse pad bytes
        try:
            s = struct.unpack_from(fmt,aset,idx)[0]
            ss.append(s)
            idx += esize
        except struct.error:
            raise error(errno.EINVAL,"Set elements failed to unpack")
    return ss

def nla_put(msg,v,a,d):
    """
     append attribute to msg's attribute list
     :param msg: GENLMsg
     :param v: attribute value
     :param a: attribute type
     :param d: attribute datatype
    """
    if d > nlh.NLA_TYPE_MAX: raise error(errno.ERANGE,"Value type is invalid")
    msg['attrs'].append((a,v,d))

# nla_put_* append data of specified datatype
def nla_put_flag(msg,a): nla_put(msg,None,a,nlh.NLA_FLAG)
def nla_put_unspec(msg,v,a): nla_put(msg,v,a,nlh.NLA_UNSPEC)
def nla_put_u8(msg,v,a): nla_put(msg,v,a,nlh.NLA_U8)
def nla_put_u16(msg,v,a): nla_put(msg,v,a,nlh.NLA_U16)
def nla_put_u32(msg,v,a): nla_put(msg,v,a,nlh.NLA_U32)
def nla_put_u64(msg,v,a): nla_put(msg,v,a,nlh.NLA_U64)
def nla_put_string(msg,v,a): nla_put(msg,v,a,nlh.NLA_STRING)
def nla_put_msecs(msg,v,a): nla_put(msg,v,a,nlh.NLA_MSECS)
def nla_put_nested(msg,v,a): nla_put(msg,v,a,nlh.NLA_NESTED)
def nla_put_set_u8(msg,v,a): nla_put(msg,v,a,nlh.NLA_SET_U8)
def nla_put_set_u16(msg,v,a): nla_put(msg,v,a,nlh.NLA_SET_U16)
def nla_put_set_u32(msg,v,a): nla_put(msg,v,a,nlh.NLA_SET_U32)
def nla_put_set_u64(msg,v,a): nla_put(msg,v,a,nlh.NLA_SET_U64)

def nla_putat(msg,i,v,a,d):
    """
     puts (overwrites) attribute at index i in msg's attr list
     :param msg: GENLMsg
     :param i: index to put attribute
     :param v: attribute value
     :param a: attribute type
     :param d: attribute datatype
    """
    if d > nlh.NLA_TYPE_MAX: raise error(errno.ERANGE,"Invalid datatype")
    msg['attrs'][i] = (a,v,d)

def nla_pop(msg,i):
    """
     pop and return the attr tuple at i in msg's attr list
     :param msg: GENLMsg
     :param i: index to pop
     :returns: the 'popped' attribute
    """
    attr = msg.attrs[i]
    del msg['attrs'][i]
    return attr

def nla_find(msg,a,value=True):
    """
     find the first attribute having type a in msg's attr list
     :param msg: GENLMsg
     :param a: attribute
     :param value: {True=return attr. value only|False=return attr. triple}
     :returns: first attribute found with a or None
    """
    for t,v,d in msg.attrs:
        if t == a:
            if value: return v
            else: return t,v,d
    return None

def nla_get(msg,i,value=True):
    """
     get attribute at index i in msg's attribute list
     :param msg: GENLMsg
     :param i: index of desired attribute
     :param value: {True=return attr. value only|False=return attr. triple}
     :returns: attribute at i
    """
    attr = msg.attrs[i]
    if value: return attr[1]
    else: return attr

#### FILE PRIVATE ####

def _nla_strip_(v):
    """
     strips padding from v
     :param v: value to strip
     :returns: v w/o padding
     **NOTE: Do not use on numeric attributes
    """
    try:
        for i,e in reversed(list(enumerate(v))):
            if e != '\x00': return v[:i+1]
        return v
    except IndexError:
        return v

def _attrpack_(a,v,d):
    """
     :param a: attribute type
     :param v: value to pack
     :param d: datatype of value
     :returns: packed attribute w/ padding if necessary
    """
    attr = "" # appease PyCharm
    if d == nlh.NLA_UNSPEC: attr = v
    elif d == nlh.NLA_U8: attr = struct.pack("B",v)
    elif d == nlh.NLA_U16: attr = struct.pack("H",v)
    elif d == nlh.NLA_U32: attr = struct.pack("I",v)
    elif d == nlh.NLA_U64: attr = struct.pack("Q",v)
    elif d == nlh.NLA_STRING:
        if _PY3_:
            # noinspection PyArgumentList
            v = bytes(v,'ascii')
        attr = struct.pack("{0}sx".format(len(v)),v)
    elif d == nlh.NLA_FLAG: attr = '' # a 0 sized attribute
    elif d == nlh.NLA_MSECS: attr = struct.pack("Q",v)
    elif d == nlh.NLA_NESTED:
        # assumes a single layer of nesting
        for nested in v:
            # prepend packed index to the already packed attribute & align it
            nattr = struct.pack('H',nested[0]) + nested[1]
            nattr += struct.pack("{0}x".format(nlh.NLMSG_ALIGNBY(len(nattr))))
            attr += nattr
    else:
        fmt = "" # appease PyCharm
        if d == nlh.NLA_SET_U8: fmt = "B"
        elif d == nlh.NLA_SET_U16: fmt = "H"
        elif d == nlh.NLA_SET_U32: fmt = "I"
        elif d == nlh.NLA_SET_U64: fmt = "Q"
        for el in v: attr += struct.pack(fmt,el)
    attr = nlh.nlattrhdr(len(attr),a) + attr
    attr += struct.pack("{0}x".format(nlh.NLMSG_ALIGNBY(len(attr))))
    return attr

def _maxbufsz_():
    """ :returns: maximum allowable socket buffer size """
    fin = None
    try:
        fin = open('/proc/sys/net/core/rmem_max')
        return int(fin.read().strip()) / 2
    except (IOError,ValueError):
        # return a hardcoded value
        return 2097152
    finally:
        if fin: fin.close()



#!/usr/bin/env python
""" wlan.py: IEEE Std 802.11-2012

Copyright (C) 2016  Dale V. Patterson (wraith.wireless@yandex.com)

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

Redistribution and use in source and binary forms, with or without modifications,
are permitted provided that the following conditions are met:
 o Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
 o Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.
 o Neither the name of the orginal author Dale V. Patterson nor the names of any
   contributors may be used to endorse or promote products derived from this
   software without specific prior written permission.

Definition of constants et al found in IEEE Std 802.11-2012

Std will refer to IEEE Std 802.11-2012
"""

__name__ = 'wlan'
__license__ = 'GPLv3'
__version__ = '0.0.1'
__date__ = 'June 2016'
__author__ = 'Dale Patterson'
__maintainer__ = 'Dale Patterson'
__email__ = 'wraith.wireless@yandex.com'
__status__ = 'Production'

"""
 cipher suite selectors - decided to (over)document this because it was such a
 hassle to figure out, want to make sure I can come back at a later date and
 figure it out again faster

 From nl80211.h @NL80211_ATTR_CIPHER_SUITES: a set of u32 values indicating the
 supported cipher suites

 The returned cipher suite (from phyinfo) for an alfa card is:
     \x01\xac\x0f\x00\x05\xac\x0f\x00\x02\xac\x0f\x00\x04\xac\x0f\x00
 which is not a nested attribute. Does 'set' mean something? There is no set
 or array or list defined netlink.h and I cannot find any reference to such in
  http://www.carisma.slowglass.com/~tgr/libnl/doc/core.html
 Another way nl80211 breaks the rules or another way I'm just not getting it?

 iw 3.17 info.c includes, but where did these values come from?
 case 0x000fac01: return "WEP40 (00-0f-ac:1)";
 case 0x000fac05: return "WEP104 (00-0f-ac:5)";
 case 0x000fac02: return "TKIP (00-0f-ac:2)";
 case 0x000fac04: return "CCMP (00-0f-ac:4)";
 case 0x000fac06: return "CMAC (00-0f-ac:6)";
 case 0x000fac08: return "GCMP (00-0f-ac:8)";
 case 0x00147201: return "WPI-SMS4 (00-14-72:1)";

 The only reference is in nl80211.h which says:
  @NL80211_KEY_CIPHER: key cipher suite (u32, as defined by IEEE 802.11 section
  7.3.2.25.1, e.g. 0x000FAC04)
 Looking in the standard we find Table 8-99 in Std which defines these values.

 Lets look in ieee80211.h and voila we find
 #define WLAN_CIPHER_SUITE_USE_GROUP	0x000FAC00
 #define WLAN_CIPHER_SUITE_WEP40		0x000FAC01
 #define WLAN_CIPHER_SUITE_TKIP		0x000FAC02
 /* reserved: 				0x000FAC03 */
 #define WLAN_CIPHER_SUITE_CCMP		0x000FAC04
 #define WLAN_CIPHER_SUITE_WEP104	0x000FAC05
 #define WLAN_CIPHER_SUITE_AES_CMAC	0x000FAC06
 #define WLAN_CIPHER_SUITE_GCMP		0x000FAC08
 #define WLAN_CIPHER_SUITE_SMS4		0x00147201

 Recall our results for the cipher key
  \x01\xac\x0f\x00\x05\xac\x0f\x00\x02\xac\x0f\x00\x04\xac\x0f\x00
 Looks similar to above. u32 is four bytes what does that give us?
  \x01\xac\x0f\x00
  \x05\xac\x0f\x00
  \x02\xac\x0f\x00
  \x04\xac\x0f\x00
 Aha, it's reversed. All we have to do is:
  hex(struct.unpack('I','\x01\xac\x0f\x00')[0])
 and we get
  '0xfac01' => WEP-40

 What a convoluted trip down the rabbit hole. This is why I hate when people
 say read the source code.

"""
#WLAN_CIPHER_SUITE_LEN = 4
WLAN_CIPHER_SUITE_GROUP    = 0x000fac00
WLAN_CIPHER_SUITE_WEP40    = 0x000fac01
WLAN_CIPHER_SUITE_TKIP     = 0x000fac02
WLAN_CIPHER_SUITE_CCMP     = 0x000fac04
WLAN_CIPHER_SUITE_WEP104   = 0x000fac05
WLAN_CIPHER_SUITE_ACS_CMAC = 0x000fac06
WLAN_CIPHER_SUITE_GCMP     = 0x000fac08
WLAN_CIPHER_SUITE_SMS4     = 0x00147201
WLAN_CIPHER_SUITE_SELECTORS = {
    WLAN_CIPHER_SUITE_GROUP:'GROUP',
    WLAN_CIPHER_SUITE_WEP40:'WEP-40',
    WLAN_CIPHER_SUITE_TKIP:'TKIP',
    WLAN_CIPHER_SUITE_CCMP:'CCMP',
    WLAN_CIPHER_SUITE_WEP104:'WEP-104',
    WLAN_CIPHER_SUITE_ACS_CMAC:'AES-CMAC',
    WLAN_CIPHER_SUITE_GCMP:'GCMP',
    WLAN_CIPHER_SUITE_SMS4:'SMS4'
}

""" COV Class Limits IAW Std Table 8-56 """
COV_CLASS_MIN =  0
COV_CLASS_MAX = 31

"""
 Retry (short and long) Limits IAW Std dot11ShortRetryLimit pg 2133 and
 dot11LongRetryLimit pg 2134
"""
RETRY_MIN = 1
RETRY_MAX = 255

""" RTS THRESH limits IAW Std dot11RTSTHRESH definition pg 2133 """
RTS_THRESH_MIN = 0
RTS_THRESH_MAX = 65536
RTS_THRESH_OFF = 4294967295 #(2^32 -1 or the max value of a u32)

""" Fragmentation THRESH limits IAW Std dot11FragmentTHRESH def. pg 2133 """
FRAG_THRESH_MIN = 256
FRAG_THRESH_MAX = 8000
FRAG_THRESH_OFF = 4294967295 #(2^32 -1 or the max value of a u32)